/*
 * Rjb - Ruby <-> Java Bridge
 * Copyright(c) 2004 arton
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * $Id:$
 * $Log:$
 */

#include <stdlib.h>
#include <stdio.h>
#include "ruby.h"
#include "util.h"
#include "jni.h"
#include "RBridge.h"

#define JVM_TYPE "client"

#if defined(_WIN32) || defined(__CYGWIN__)
 #if defined(__CYGWIN__)
  #define JVMDLL "%s/jre/bin/%s/jvm.dll"
  #define DIRSEPARATOR '/'
 #else
  #define JVMDLL "%s\\jre\\bin\\%s\\jvm.dll"
  #define DIRSEPARATOR '\\'
 #endif
#else /* defined(_WIN32) || defined(__CYGWIN__) */
 #ifdef _LP64
  #ifdef ia64
   #define ARCH "ia64"
  #else
   #define ARCH "sparcv9"
  #endif
 #endif
 #if defined(i586) || defined(__i386__)
  #define ARCH "i386"
 #endif
 #ifdef __sparc
  #include <sys/systeminfo.h>
 #endif
 #ifndef ARCH
  #include <sys/systeminfo.h>
 #endif
 #define JVMDLL "%s/jre/lib/%s/%s/libjvm.so"
 #define DIRSEPARATOR '/'
#endif

typedef void (*GETDEFAULTJAVAVMINITARGS)(void*);
typedef int (*CREATEJAVAVM)(JavaVM**, void**, void*);

extern void check_exception(int);

extern JavaVM* jvm;
extern JNIEnv* jenv;
extern jclass rbridge;
extern jmethodID register_bridge;
static VALUE jvmdll = Qnil;
static VALUE GetDefaultJavaVMInitArgs;
static VALUE CreateJavaVM;

/*
 * not completed, only valid under some circumstances.
 */
static VALUE load_jvm(char* jvmtype)
{
    char* libpath;
    char* java_home;
    char* jh;
    VALUE dl;

    jh = getenv("JAVA_HOME");
    if (!jh)
    {
        return Qnil;
    }
    java_home = ALLOCA_N(char, strlen(jh) + 1);
    strcpy(java_home, jh);
    if (*(java_home + strlen(jh) - 1) == DIRSEPARATOR)
    {
	*(java_home + strlen(jh) - 1) = '\0';
    }
#if defined(_WIN32) || defined(__CYGWIN__)
    libpath = ALLOCA_N(char, sizeof(JVMDLL) + strlen(java_home)
		       + strlen(jvmtype) + 1);
    sprintf(libpath, JVMDLL, java_home, jvmtype);
#else /* not Windows */
    libpath = ALLOCA_N(char, sizeof(JVMDLL) + strlen(java_home)
		       + strlen(ARCH) + strlen(jvmtype) + 1);
    sprintf(libpath, JVMDLL, java_home, ARCH, jvmtype);
#endif
    rb_require("dl");
    dl = rb_eval_string("DL");
    jvmdll = rb_funcall(dl, rb_intern("dlopen"), 1, rb_str_new2(libpath));
    GetDefaultJavaVMInitArgs = rb_funcall(jvmdll, rb_intern("sym"),
			  2, rb_str_new2("JNI_GetDefaultJavaVMInitArgs"),
			     rb_str_new2("0p"));
    CreateJavaVM = rb_funcall(jvmdll, rb_intern("sym"),
			      2, rb_str_new2("JNI_CreateJavaVM"),
			         rb_str_new2("pppp"));
    return Qtrue;
}

int load_bridge()
{
    JNINativeMethod nmethod[1];
    jbyte buff[8192];
    jclass loader = (*jenv)->FindClass(jenv, "java/lang/ClassLoader");
    jmethodID getSysLoader = (*jenv)->GetStaticMethodID(jenv, loader,
		   "getSystemClassLoader", "()Ljava/lang/ClassLoader;");
    jobject iloader = (*jenv)->CallStaticObjectMethod(jenv, loader, getSysLoader);
    FILE* f = fopen("RBridge.class", "rb");
    int len = fread(buff, 1, sizeof(buff), f);
    fclose(f);
    rbridge = (*jenv)->DefineClass(jenv,
		   "jp/co/infoseek/hp/arton/rjb/RBridge", iloader, buff, len);
    if (rbridge == NULL)
    {
	check_exception(1);
    }
    register_bridge = (*jenv)->GetMethodID(jenv, rbridge, "register",
			   "(Ljava/lang/Class;)Ljava/lang/Object;");
    nmethod[0].name = "call";
    nmethod[0].signature = "(Ljava/lang/String;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;";
    nmethod[0].fnPtr = Java_RBridge_call;
    (*jenv)->RegisterNatives(jenv, rbridge, nmethod, 1);
    return 0;
}

int create_jvm(JDK1_1InitArgs* vm_args, char* userpath)
{
    VALUE ptr;
    int len;
    int result;
    GETDEFAULTJAVAVMINITARGS initargs;
    CREATEJAVAVM createjavavm;
    if (jvmdll == Qnil)
    {
	if (load_jvm(JVM_TYPE) == Qnil)
	{
	    return -1;
	}
    }

#if defined(_WIN32) || defined(__CYGWIN__)
    ptr = rb_funcall(GetDefaultJavaVMInitArgs, rb_intern("to_i"), 0);
    initargs = *(GETDEFAULTJAVAVMINITARGS*)FIX2INT(ptr);
    initargs(vm_args);
#endif

    len = strlen(userpath);
    if (vm_args->classpath)
    {
	char* newpath = ALLOCA_N(char, len + strlen(vm_args->classpath) + 2);
	sprintf(newpath, "%s%c%s", userpath, PATH_SEP, vm_args->classpath);
	vm_args->classpath = newpath;
    }
    else
    {
	vm_args->classpath = userpath;
    }
    ptr = rb_funcall(CreateJavaVM, rb_intern("to_i"), 0);
    createjavavm = *(CREATEJAVAVM*)FIX2INT(ptr);
    result = createjavavm(&jvm, (void**)&jenv, vm_args);
    if (!result)
    {
	result = load_bridge();
    }
    return result;
}

